from typing import Iterable, Dict, Any, List, Set, Tuple


def combine_dicts(dicts: Iterable[Dict[str, Any]]) -> Tuple[List[Dict[str, Any]], Set[str]]:
    """
    Combines a series of ROI-format dictionaries into a list of
    dictionaries with a common set of headers.

    :param dicts:   The dictionaries to combine.
    :return:        The combined dictionaries, and the set of headers for each dictionary.
    """
    # Cache the ROI dicts
    dicts = list(dicts)

    # Combine the headers
    headers = set()
    for roi_dict in dicts:
        headers.update(roi_dict.keys())

    # Add missing headers to each dict
    for roi_dict in dicts:
        for header in headers:
            if header not in roi_dict:
                roi_dict[header] = ""

    return dicts, headers



import os
import re
from typing import Optional, Tuple, List

from wai.annotations.domain.image import ImageFormat
from wai.annotations.domain.image.util import get_associated_image
from ..constants import DEFAULT_EXTENSION


def roi_filename_for_image(image_filename: str,
                           prefix: Optional[str] = None,
                           suffix: Optional[str] = None) -> str:
    """
    Creates a filename for the ROI CSV file from the given image filename.

    :param image_filename:  The image filename.
    :param prefix:          The optional prefix to the filename.
    :param suffix:          The optional suffix to the filename.
    :return:                The ROI CSV filename.
    """
    # Set defaults for prefix and suffix
    prefix, suffix = handle_fix_defaults(prefix, suffix)

    return f"{prefix}{os.path.splitext(image_filename)[0]}{suffix}"


def get_associated_image_from_filename(roi_filename: str,
                                       prefix: Optional[str] = None,
                                       suffix: Optional[str] = None,
                                       preference_order: Optional[List[ImageFormat]] = None) -> str:
    """
    Gets the associated image filename for a given ROI filename.

    :param roi_filename:        The filename of the ROI file.
    :param prefix:              The prefix for the ROI filename.
    :param suffix:              The suffix for the ROI filename.
    :param preference_order:    The preferred format order in which to search for images.
    :return:                    The image filename.
    """
    # Get the necessary fix defaults
    prefix, suffix = handle_fix_defaults(prefix, suffix)

    # Create a regex matcher to extract the image basename from the ROI filename
    matcher = re.compile(f"(.*){prefix}(.*){suffix}$")

    # Match the ROI filename
    match = matcher.match(roi_filename)

    # Create the base image filename
    image_file = match.group(1) + match.group(2)

    # Find the image file that matches this name
    image_file = get_associated_image(image_file, preference_order)

    # Make sure an associated image was found
    if image_file is None:
        raise ValueError(f"No associated image found for {roi_filename}")

    return os.path.basename(image_file)


def handle_fix_defaults(prefix: Optional[str] = None,
                        suffix: Optional[str] = None) -> Tuple[str, str]:
    """
    Returns the prefix and suffix values, using defaults where
    values aren't supplied.

    :param prefix:  The prefix, or None to get the default.
    :param suffix:  The suffix, or None to get the default.
    :return:        The prefix and suffix to use.
    """
    # Get the defaults where necessary
    prefix = prefix if prefix is not None else ""
    suffix = suffix if suffix is not None else DEFAULT_EXTENSION

    return prefix, suffix



"""
Module defining the ROI external format.
"""
from typing import Tuple, List

from wai.annotations.domain.image import Image
from ._ROIObject import ROIObject


# Image info, ROI annotations
ROIExternalFormat = Tuple[Image, List[ROIObject]]



from typing import List, Tuple

import numpy as np

from wai.annotations.image_utils import polygon_to_minrect


def min_rect_from_roi_polygon(poly_x: List[float], poly_y: List[float]) -> Tuple[float, float]:
    """
    Gets the width and height of the minimum-area rectangle around
    a ROI object's polygon.

    :param poly_x:  The list of x coordinates of the polygon.
    :param poly_y:  The list of y coordinates of the polygon.
    :return:        The width and height of the minimum-area rectangle.
    """
    # Create the array expected by polygon_to_minrect
    poly_array = np.array(list(zip(poly_x, poly_y)), dtype=np.float32)

    return polygon_to_minrect(poly_array)



import os
import csv

from wai.common.cli.options import TypedOption

from wai.annotations.core.component.util import AnnotationFileProcessor
from wai.annotations.core.stream import ThenFunction
from wai.annotations.domain.image import Image, ImageFormat
from ..util import ROIExternalFormat, ROIObject, get_associated_image_from_filename, consume_comments


class ROIReader(AnnotationFileProcessor[ROIExternalFormat]):
    """
    Reader of ROI-format CSV files.
    """
    # The order of preference of image formats
    image_format_preference_order = TypedOption(
        "-e", "--extensions",
        type=ImageFormat,
        nargs=3,
        metavar="FORMAT",
        default=[ImageFormat.PNG, ImageFormat.JPG, ImageFormat.BMP],
        help="image format extensions in order of preference"
    )

    reader_prefix = TypedOption(
        "--prefix",
        type=str,
        help="the prefix for output filenames (default = '')"
    )

    reader_suffix = TypedOption(
        "--suffix",
        type=str,
        help="the suffix for output filenames (default = '-rois.csv')"
    )

    def read_annotation_file(self, filename: str, then: ThenFunction[ROIExternalFormat]):
        # Read in the file
        with open(filename, "r") as file:
            # Consume the comments lines
            consume_comments(file)

            # Create a CSV dict reader around the file
            csv_reader = csv.DictReader(file)

            # Extract the contents into dictionaries
            roi_dicts = [dict(row) for row in csv_reader]

        # Get the image filename
        if len(roi_dicts) > 0:
            image_file = roi_dicts[0]["file"]
        else:
            image_file = get_associated_image_from_filename(filename,
                                                            self.reader_prefix,
                                                            self.reader_suffix,
                                                            self.image_format_preference_order)

        # Get the full path to the image
        image_path = os.path.join(os.path.dirname(filename), image_file)

        # Create the ROI objects from the dictionaries
        roi_objects = list(map(ROIObject.from_dict, roi_dicts))

        then((Image.from_file(image_path), roi_objects))

    def read_negative_file(self, filename: str, then: ThenFunction[ROIExternalFormat]):
        then((Image.from_file(filename), []))



from typing import List, Tuple

from wai.common.geometry import Polygon


def roi_polygon(polygon: Polygon) -> Tuple[List[float], List[float]]:
    """
    Converts a polygon into lists of x and y coordinates.

    :param polygon: The polygon.
    :return:        The lists of x and y coordinates.
    """
    # Create the lists
    poly_x, poly_y = [], []

    # Add each point
    for point in polygon:
        poly_x.append(round(point.x))
        poly_y.append(round(point.y))

    return poly_x, poly_y



from typing import IO

from ..constants import COMMENT_SYMBOL


def consume_comments(file: IO[str]):
    """
    Reads lines from the file until a non-comment line is found.

    :param file:    The file to read from.
    """
    # Read until end of file
    line = None
    while line != '':
        # Record the position in the file so we can backtrack
        # if we find a non-comment line
        position = file.tell()

        # Read the next line from the file
        line = file.readline()

        # If it's not a comment, backtrack and return
        if not line.startswith(COMMENT_SYMBOL):
            file.seek(position)
            return



from wai.common.adams.imaging.locateobjects import LocatedObjects, LocatedObject

from wai.annotations.core.component import ProcessorComponent
from wai.annotations.core.stream import ThenFunction, DoneFunction
from wai.annotations.core.stream.util import RequiresNoFinalisation
from wai.annotations.domain.image.object_detection import ImageObjectDetectionInstance
from wai.annotations.domain.image.object_detection.util import set_object_label, set_object_metadata, set_object_prefix
from ..util import ROIExternalFormat, ROIObject


class FromROI(
    RequiresNoFinalisation,
    ProcessorComponent[ROIExternalFormat, ImageObjectDetectionInstance]
):
    """
    Converter from ROI CSV annotations to internal format.
    """
    def process_element(
            self,
            element: ROIExternalFormat,
            then: ThenFunction[ImageObjectDetectionInstance],
            done: DoneFunction
    ):
        image_info, roi_objects = element

        then(
            ImageObjectDetectionInstance(
                image_info,
                LocatedObjects(map(self.convert_roi_object, roi_objects))
            )
        )

    @staticmethod
    def convert_roi_object(roi_object: ROIObject) -> LocatedObject:
        """
        Converts a ROI object to an internal-format located object.

        :param roi_object:  The ROI object to convert.
        :return:            The located object.
        """
        located_object = LocatedObject(round(roi_object.x),
                                       round(roi_object.y),
                                       round(roi_object.w),
                                       round(roi_object.h))

        set_object_label(located_object, roi_object.label_str)

        # Set any non-standard keywords as meta-data
        metadata = roi_object.non_standard_kwargs.copy()
        if "prefix" in metadata:
            set_object_prefix(located_object, metadata["prefix"])
            del metadata["prefix"]
        set_object_metadata(located_object, **metadata)

        # store additional values in meta-data
        located_object.metadata["score"] = roi_object.score
        located_object.metadata["label_index"] = roi_object.label
        if hasattr(located_object, "minrect_w"):
            located_object.metadata["minrect_w"] = roi_object.minrect_w
        if hasattr(located_object, "minrect_h"):
            located_object.metadata["minrect_h"] = roi_object.minrect_h

        # Add the polygon
        if roi_object.has_polygon():
            located_object.set_polygon(roi_object.polygon())

        return located_object



from typing import Dict

from wai.common.adams.imaging.locateobjects import LocatedObject
from wai.common.cli.options import TypedOption

from wai.annotations.core.component import ProcessorComponent
from wai.annotations.core.stream import ThenFunction, DoneFunction
from wai.annotations.core.stream.util import ProcessState, RequiresNoFinalisation
from wai.annotations.domain.image.object_detection import ImageObjectDetectionInstance
from wai.annotations.domain.image.object_detection.util import get_object_label, get_object_prefix, get_object_metadata
from ..util import min_rect_from_roi_polygon, roi_polygon, ROIExternalFormat, ROIObject


class ToROI(
    RequiresNoFinalisation,
    ProcessorComponent[ImageObjectDetectionInstance, ROIExternalFormat]
):
    """
    Converter from internal format to ROI annotations.
    """
    image_size = TypedOption(
        "-d", "--image-dimensions",
        type=int,
        nargs=2,
        metavar=("WIDTH", "HEIGHT"),
        help="image dimensions to use if none can be inferred"
    )

    _label_map: Dict[str, int] = ProcessState(lambda self: {})

    def process_element(
            self,
            element: ImageObjectDetectionInstance,
            then: ThenFunction[ROIExternalFormat],
            done: DoneFunction
    ):
        image_info, located_objects = element

        # Make sure we can get the image size
        if image_info.size is not None:
            image_size = image_info.size
        elif self.image_size is not None:
            image_size = self.image_size
        else:
            raise RuntimeError(f"Cannot infer image dimensions for {image_info.filename}")

        # Create a wrapper to convert located objects using the image size
        def convert(located_object: LocatedObject) -> ROIObject:
            return self.convert_located_object(located_object, *image_size)

        # If this is a negative instance, create an empty iterable
        if located_objects is None:
            located_objects = tuple()

        then((image_info, list(map(convert, located_objects))))

    def convert_located_object(self,
                               located_object: LocatedObject,
                               image_width: int,
                               image_height: int) -> ROIObject:
        """
        Converts a located object to ROI format.

        :param located_object:  The located object.
        :param image_width:     The width of the image.
        :param image_height:    The height of the image.
        :return:                The ROI object.
        """
        # Get the object's rectangle
        rectangle = located_object.get_rectangle()

        # Get the object's label
        label = get_object_label(located_object)

        # Add the label to the label map if it's not already present
        if label not in self._label_map:
            self._label_map[label] = len(self._label_map)

        # Create keyword arguments from the required attributes
        kwargs = dict(x0=float(rectangle.left()),
                      y0=float(rectangle.top()),
                      x1=float(rectangle.right()),
                      y1=float(rectangle.bottom()),
                      x0n=rectangle.left() / image_width,
                      y0n=rectangle.top() / image_height,
                      x1n=rectangle.right() / image_width,
                      y1n=rectangle.bottom() / image_height,
                      label=self._label_map[label],
                      label_str=label,
                      score=1.0)

        # If there is a polygon, add them as keyword arguments
        if located_object.has_polygon():
            poly_x, poly_y = roi_polygon(located_object.get_polygon())
            minrect_w, minrect_h = min_rect_from_roi_polygon(poly_x, poly_y)
            kwargs.update(poly_x=poly_x,
                          poly_y=poly_y,
                          poly_xn=list(map(lambda x: x / image_width, poly_x)),
                          poly_yn=list(map(lambda y: y / image_height, poly_y)),
                          minrect_w=minrect_w,
                          minrect_h=minrect_h)

        # Set the non-standard keywords
        prefix = get_object_prefix(located_object, None)
        if prefix is not None:
            kwargs.update(prefix=prefix)
        kwargs.update(**get_object_metadata(located_object))

        return ROIObject(**kwargs)



from typing import Dict, List, Optional

from wai.common.geometry import Polygon, Point

from wai.annotations.core.logging import LoggingEnabled
from ._normalisation import normalisation_constant_multi


class ROIObject(LoggingEnabled):
    """
    Holds a detected object from a ROI file.
    """
    # The ordered required position keywords for corner mode
    required_position_keywords_corner_mode = ("x0", "y0", "x1", "y1", "x0n", "y0n", "x1n", "y1n")

    # The ordered required position keywords for corner mode
    required_position_keywords_size_mode = ("x", "y", "w", "h", "xn", "yn", "wn", "hn")

    # The ordered required arguments to ROIObject
    required_other_keywords = ("label", "label_str", "score")

    # The ordered optional arguments to ROIObject
    optional_keywords = ("poly_x", "poly_y", "poly_xn", "poly_yn", "minrect_w", "minrect_h")

    def __init__(self,
                 x0: float, y0: float, x1: float, y1: float,
                 x0n: float, y0n: float, x1n: float, y1n: float,
                 label: int, label_str: str, score: float,
                 poly_x: Optional[List[float]] = None, poly_y: Optional[List[float]] = None,
                 poly_xn: Optional[List[float]] = None, poly_yn: Optional[List[float]] = None,
                 minrect_w: Optional[float] = None, minrect_h: Optional[float] = None,
                 **non_standard_kwargs):
        self.x0: float = x0
        self.y0: float = y0
        self.x1: float = x1
        self.y1: float = y1
        self.x0n: float = x0n
        self.y0n: float = y0n
        self.x1n: float = x1n
        self.y1n: float = y1n
        self.label: int = label
        self.label_str: str = label_str
        self.score: float = score
        self.poly_x: Optional[List[float]] = poly_x
        self.poly_y: Optional[List[float]] = poly_y
        self.poly_xn: Optional[List[float]] = poly_xn
        self.poly_yn: Optional[List[float]] = poly_yn
        self.minrect_w: Optional[float] = minrect_w
        self.minrect_h: Optional[float] = minrect_h

        # Capture any additional columns
        self.non_standard_kwargs = non_standard_kwargs

        # "file" is a reserved keyword for the ROI file format
        if "file" in non_standard_kwargs:
            raise NameError(f"'file' is reserved for the ROI file format, can't be in the non-standard keywords")

        # Make sure all of the polygon lists are the same length
        self.ensure_polygon_list_lengths()

    @property
    def x(self) -> float:
        return self.x0

    @property
    def y(self) -> float:
        return self.y0

    @property
    def w(self) -> float:
        return self.x1 - self.x0 + 1

    @property
    def h(self) -> float:
        return self.y1 - self.y0 + 1

    @property
    def xn(self) -> float:
        return self.x / self.image_width()

    @property
    def yn(self) -> float:
        return self.y / self.image_height()

    @property
    def wn(self) -> float:
        return self.w / self.image_width()

    @property
    def hn(self) -> float:
        return self.h / self.image_height()

    def ensure_polygon_list_lengths(self):
        """
        Makes sure all polygon lists match in length.
        """
        # Get the length of each list
        len_x = -1 if self.poly_x is None else len(self.poly_x)
        len_y = -1 if self.poly_y is None else len(self.poly_y)
        len_xn = -1 if self.poly_xn is None else len(self.poly_xn)
        len_yn = -1 if self.poly_yn is None else len(self.poly_yn)

        # Make sure all lengths are the same
        if len_x != len_y or len_y != len_xn or len_xn != len_yn:
            raise ValueError("Polygon lists have differing lengths")

    def has_polygon(self) -> bool:
        """
        Whether this ROI object has a polygon.

        :return:    True if the object has a polygon.
        """
        return self.poly_x is not None

    def polygon(self) -> Polygon:
        """
        Generates a polygon from the given ROI object.

        :return:            The polygon.
        """
        # Make sure the ROI object has a polygon
        if not self.has_polygon():
            raise ValueError(f"ROI object has no polygon")

        return Polygon(*(Point(round(x), round(y)) for x, y in zip(self.poly_x, self.poly_y)))

    def image_width(self) -> int:
        """
        Calculates the source image's width.

        :return:    The source image's width.
        """
        return normalisation_constant_multi(self.x0, self.x0n, self.x1, self.x1n)

    def image_height(self) -> int:
        """
        Calculates the source image's height.

        :return:    The source image's height.
        """
        return normalisation_constant_multi(self.y0, self.y0n, self.y1, self.y1n)

    @classmethod
    def required_keywords(cls, size_mode: bool):
        return (
                (cls.required_position_keywords_size_mode if size_mode else cls.required_position_keywords_corner_mode)
                + cls.required_other_keywords
        )

    @classmethod
    def keywords(cls, size_mode: bool):
        return cls.required_keywords(size_mode) + cls.optional_keywords

    def as_dict(self, size_mode: bool = False) -> Dict[str, str]:
        """
        Gets a dictionary representation of this object.

        :param size_mode:   Whether to export as a size-mode dictionary instead.
        :return:            A dict.
        """
        # Create a result dict
        result = {}

        # Get the keywords we expect to export
        keywords = self.keywords(size_mode)

        # Process each keyword in turn
        for keyword in keywords:
            # Get the value for the attribute
            value = getattr(self, keyword)

            # Ignore missing optional attributes
            if value is None:
                continue

            # Format the polygon lists as strings
            if isinstance(value, list):
                value = ",".join(map(str, value))

            # Add the string representation of the value to the dict
            result[keyword] = value

        result.update(**self.non_standard_kwargs)

        return result

    @classmethod
    def from_dict(cls, roi_dict: Dict[str, str]) -> "ROIObject":
        """
        Creates an object instance from the given dictionary.

        :param roi_dict:    The dict.
        :return:            A ROIObject instance.
        """
        # Create a set of keyword arguments
        kwargs = {}

        # Process each dictionary item in turn
        for keyword, value in roi_dict.items():
            # Ignore the file keyword
            if keyword == "file":
                continue

            # Add the value to the kwargs
            kwargs[keyword] = cls._parse(keyword, value)

        # If it's a size-mode dict, convert to a corner-mode one
        if "x" in kwargs:
            kwargs = cls._convert_size_mode_kwargs(**kwargs)

        return ROIObject(**kwargs)

    @classmethod
    def _parse(cls, keyword: str, value: str):
        """
        Parses the string representation of a value to its internal type.

        :param keyword: The name of the attribute.
        :param value:   The string representation of the attribute's value.
        """
        # Float types
        if keyword in {"x0", "y0", "x1", "y1", "x0n", "y0n", "x1n", "y1n",
                       "x", "y", "w", "h", "xn", "yn", "wn", "hn",
                       "score", "minrect_w", "minrect_h"}:
            return float(value) if value is not "" else None

        # Int types
        if keyword in {"label"}:
            return int(value)

        # List of float types
        if keyword in {"poly_x", "poly_y", "poly_xn", "poly_yn"}:
            return list(map(float, value.split(","))) if value is not "" else None

        # All other attributes are strings
        return value

    @classmethod
    def _convert_size_mode_kwargs(cls, **kwargs) -> Dict:
        """
        Converts the keyword arguments to the constructor from
        size-mode to the equivalent corner-mode dictionary.

        :param kwargs:  The size-mode keyword arguments.
        :return:        The corner-mode keyword arguments.
        """
        # Remove the size-mode arguments
        size_mode_kwargs = {}
        for keyword in cls.required_position_keywords_size_mode:
            size_mode_kwargs[keyword] = kwargs.pop(keyword)

        # Calculate the image's width/height for normalisation
        image_width = normalisation_constant_multi(size_mode_kwargs['x'],
                                                   size_mode_kwargs['xn'],
                                                   size_mode_kwargs['w'],
                                                   size_mode_kwargs['wn'])
        image_height = normalisation_constant_multi(size_mode_kwargs['y'],
                                                    size_mode_kwargs['yn'],
                                                    size_mode_kwargs['h'],
                                                    size_mode_kwargs['hn'])

        # Introduce the corner-mode equivalent arguments
        kwargs['x0'] = size_mode_kwargs['x']
        kwargs['y0'] = size_mode_kwargs['y']
        kwargs['x1'] = size_mode_kwargs['x'] + size_mode_kwargs['w'] - 1
        kwargs['y1'] = size_mode_kwargs['y'] + size_mode_kwargs['h'] - 1
        kwargs['x0n'] = kwargs['x0'] / image_width
        kwargs['y0n'] = kwargs['y0'] / image_height
        kwargs['x1n'] = kwargs['x1'] / image_width
        kwargs['y1n'] = kwargs['y1'] / image_height

        return kwargs



import os
from typing import IO
import csv

from wai.common.cli.options import TypedOption, FlagOption

from wai.annotations.core.component.util import (
    SplitSink,
    SeparateFileWriter,
    RequiresNoSplitFinalisation,
    ExpectsDirectory
)
from ..constants import COMMENT_SYMBOL
from ..util import combine_dicts, roi_filename_for_image, ROIExternalFormat, ROIObject


class ROIWriter(
    ExpectsDirectory,
    RequiresNoSplitFinalisation,
    SplitSink[ROIExternalFormat], SeparateFileWriter[ROIExternalFormat]
):
    """
    Writer of ROI CSV annotations.
    """
    writer_prefix = TypedOption(
        "--prefix",
        type=str,
        help="the prefix for output filenames (default = '')"
    )

    writer_suffix = TypedOption(
        "--suffix",
        type=str,
        help="the suffix for output filenames (default = '-rois.csv')"
    )

    comments = TypedOption(
        "--comments",
        type=str,
        nargs="+",
        help="comments to write to the beginning of the ROI file"
    )

    size_mode = FlagOption(
        "--size-mode",
        help="writes the ROI files with x,y,w,h headers instead of x0,y0,x1,y1"
    )

    def start(self):
        super().start()

        # Path must be a directory
        if not os.path.isdir(self.output):
            raise ValueError(f"{self.output} is not a directory, or does not exist")

    def consume_element_for_split(
            self,
            element: ROIExternalFormat
    ):
        # Unpack the instance
        image_info, roi_objects = element

        # Format the report filename
        filename: str = roi_filename_for_image(image_info.filename, self.writer_prefix, self.writer_suffix)

        # Format the filename for the ROI file
        roi_filename = self.get_split_path(self.split_label, os.path.join(self.output, filename), True)

        # Write the image
        self.write_data_file(image_info, os.path.dirname(roi_filename))

        # If this is a negative example, we're done
        if len(roi_objects) == 0:
            return

        # Format each ROI object as a dictionary
        roi_dicts, headers = combine_dicts(roi_object.as_dict(self.size_mode) for roi_object in roi_objects)

        # Get the header keywords we are expecting
        keywords = ROIObject.keywords(self.size_mode)

        # Extract the non-standard headers
        non_standard_headers = headers - set(keywords)

        # Put the standard headers in order
        headers = tuple(header for header in keywords
                        if header in headers or header in set(ROIObject.required_keywords(self.size_mode)))

        # Add the filename header
        headers = "file", *headers, *non_standard_headers

        # Write the CSV file
        with open(roi_filename, "w") as file:
            self._write_comments(file)
            csv_writer = csv.DictWriter(file, headers)
            csv_writer.writeheader()
            for roi_dict in roi_dicts:
                roi_dict.update(file=image_info.filename)
                csv_writer.writerow(roi_dict)

    @classmethod
    def get_help_text_for_output_option(cls) -> str:
        return "output directory to write files to"

    def _write_comments(self, file: IO[str]):
        """
        Writes the supplied comments to the given file.

        :param file:    The file to write to.
        """
        # Write each entry on its own line
        for comment_line in self.comments:
            # Add the comment symbol if it's not there already
            if not comment_line.startswith(COMMENT_SYMBOL):
                comment_line = f"{COMMENT_SYMBOL} {comment_line}"

            # Write the comment
            file.write(f"{comment_line}\n")



"""
Module for constants relating to the ROI CSV format.
"""
# The default extension for ROI CSV files
DEFAULT_EXTENSION = "-rois.csv"

# The symbol indicating a line in the ROI file is a comment
COMMENT_SYMBOL = "#"



"""
Module for functions relating to normalisation of values.
"""


def normalisation_constant(unnormalised: float, normalised: float) -> int:
    """
    Returns the constant used to normalise a value.

    :param unnormalised:    The unnormalised value.
    :param normalised:      The normalised value.
    :return:                The normalisation constant.
    """
    return round(unnormalised / normalised)


def normalisation_constant_multi(*values: float):
    """
    Calculates the normalisation constant from multiple pairs
    of unnormalised/normalised values, using the first pair
    where the normalised value is not zero.

    :param values:  The value pairs.
    :return:        The normalisation constant.
    """
    # Make sure we have at least one pair
    if len(values) < 2:
        raise ValueError("Must provide at least two values to calculate normalisation constant")

    # Check each pair except the last
    for unnormalised, normalised in zip(values[0:-2:2], values[1:-1:2]):
        if normalised != 0.0:
            return normalisation_constant(unnormalised, normalised)

    # If all other pairs fail, use the last pair
    return normalisation_constant(values[-2], values[-1])



"""
Package of utilities for working with the ROI format.
"""
from ._combine_dicts import combine_dicts
from ._consume_comments import consume_comments
from ._format import ROIExternalFormat
from ._min_rect_from_roi_polygon import min_rect_from_roi_polygon
from ._normalisation import normalisation_constant, normalisation_constant_multi
from ._filenames import (
    roi_filename_for_image,
    get_associated_image_from_filename,
    handle_fix_defaults
)
from ._roi_polygon import roi_polygon
from ._ROIObject import ROIObject



from typing import Type, Tuple

from wai.annotations.core.component import Component
from wai.annotations.core.domain import DomainSpecifier
from wai.annotations.core.specifier import SinkStageSpecifier


class ROIOutputFormatSpecifier(SinkStageSpecifier):
    """
    Specifier of the components for writing the ROI CSV-based
    object detection format.
    """
    @classmethod
    def description(cls) -> str:
        return "Writes image object-detection annotations in the ROI CSV-format"

    @classmethod
    def components(cls) -> Tuple[Type[Component], ...]:
        from ..component import ToROI, ROIWriter
        return ToROI, ROIWriter

    @classmethod
    def domain(cls) -> Type[DomainSpecifier]:
        from wai.annotations.domain.image.object_detection import ImageObjectDetectionDomainSpecifier
        return ImageObjectDetectionDomainSpecifier
